/* $Id: utility.c,v 1.11 2000/08/07 04:57:15 kouno Exp $ */

/* 
 *  Project: IPMSG Utility
 *  Author & Copyright:  KOHNO TAKASHI(δ) <kouno@catnet.ne.jp>
 *  Mail: kouno@catnet.ne.jp
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>		   
#include <netdb.h>
#include <signal.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <ctype.h>

#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "protocol.h"
#include "tipmsg.h"

#if defined(DEBUG)
#define line	fprintf(stderr, "Line %d\n", __LINE__);fflush(stderr);fflush(stdout)
#define d_printf printf
#else
#define debug
#define d_printf dummy_function
#endif

/* ۥ̾ -> ɥ쥹 Ѵ */
long hosttoaddr(char *target)
{
	unsigned long addr;
	struct hostent *host;

    if (strcmp(target, BROADCAST) == 0){
        return ~0;
    }

	if ((addr = inet_addr(target)) == -1){
		if ((host = (struct hostent *)gethostbyname(target)) != NULL){
			addr = *(long *)host->h_addr;
		} else {
            fprintf(stderr, "invalid hostname %s\n", target);
			perror(NULL);
			return 0;
		}
	}
    d_printf("Get host addr : %s -> %s\n", target, Xinet_ntoa(addr));
	return addr;
}

/* ۥ̾ -> ɥ쥹 Ѵ (桼ꥹȤ)*/
long ihosttoaddr(char *target, USERLIST_MNG* ulm)
{
	USERLIST* ul;
	if (!(ul = search_userlist_by_host(ulm, target))){
		d_printf("Not found host in userlist, step in hosttoaddr.\n");
		return hosttoaddr(target);
	}
    
	return ul->addr;
}

/* 桼̾ -> ɥ쥹 Ѵ (桼ꥹȤ)*/
long iusertoaddr(char *target, USERLIST_MNG* ulm)
{
	USERLIST* ul;
	if (!(ul = search_userlist_by_user(ulm, target))){
		d_printf("Not found user in userlist, step in usertoaddr.\n");
		return 0;
	}
    
	return ul->addr;
}

/* ХåեΥå¤ΤؤŸ */
int decom_msg(MSG *msg, const char *buf, int len)
{
	const char* token = buf;
	char* endp;
	int lp;

	/* Get format version */
	msg->format_ver = strtol(token, &endp, 10);
	if (msg->format_ver != IPMSG_VERSION && msg->format_ver != TIPMSG_CMD_VERSION){
		return 0;
	}
	if (!endp || *endp != ':'){
		fprintf(stderr, "Can't find format version\n" );
		if (!(endp = strchr(token, ':'))) return 0;
	}
	token = endp+1;
	
    /* Proxy working */
	if (msg->format_ver == TIPMSG_CMD_VERSION){
		msg->cmd_ver = msg->format_ver;
		/* TIPMSG Command Aera */
		msg->cmd = strtol(token, &endp, 10);
		if (!endp || *endp != ':'){
			fprintf(stderr, "Can't find cmd\n");
			if (!(endp = strchr(token, ':'))) return 0;
		}
		token = endp+1;
		/* TIPMSG Data Aera */
		for (lp=0; token[lp] != ':'; lp++){
			if (token[lp] == '\0'){
				fprintf(stderr, "Can't find tipmsg data\n" );
				return 0;
			}
			msg->data[lp] = token[lp];
		}
		msg->data[lp] = '\0';
		token+=lp+1;
		d_printf("TIPMSG CMD: 0x%x:%s\n", msg->cmd, msg->data);
		return decom_msg(msg, token, len - (token - buf));
	}

	/* Get Packet No */
	msg->packet_no=strtol(token, &endp, 10);
	if (!endp || *endp != ':'){
		fprintf(stderr, "Can't find Packet\n");
		if (!(endp = strchr(token, ':'))) return 0;
	}
	token = endp+1;
	
	/* Get Sender Name */
	for (lp=0; token[lp] != ':'; lp++){
		if (token[lp] == '\0'){
			fprintf(stderr, "Can't find Sender Name\n" );
			return 0;
		}
		msg->msg_user[lp] = token[lp];
	}
	msg->msg_user[lp] = '\0';
	token+=lp+1;

	/* Get Sendhost Name */
	for (lp=0; token[lp] != ':'; lp++){
		if (token[lp] == '\0'){
			fprintf(stderr, "Can't find Sendhost Name\n" );
			return 0;
		}
		msg->msg_host[lp] = token[lp];
	}
	msg->msg_host[lp] = '\0';
	token+=lp+1;
	
	/* Get command */
	msg->command=strtol(token, &endp, 10) ;
	if (!endp || *endp != ':'){
		fprintf(stderr, "Can't find command\n" );
		if (!(endp = strchr(token, ':'))) return 0;
	}
	token = endp+1;

	/* Get Extension */
	msg->len = len - (token - buf);
	for (len = 0; token[len] != '\0' && len < msg->len; len++);
	memcpy(msg->extend, token, msg->len);
	msg->group = msg->extend + len +1;
	if (len >= msg->len){
		/* No groupname presented */
		msg->group --; 
	}
	
	return 1;
}

/* Message ¸ */
int store_message(MSG* msg, struct sockaddr_in* psaddr, 
				  CONFIG_MNG* cfg, USERLIST_MNG* ulm)
{
    FILE* fp;
    static int count = 0;
    char file[PATH_MAX];
    static char master[PATH_MAX]="";
    char *endp;
    char *startp;
	int len;

    fp = stdout;

	/* åֹν */
    if (!count && cfg->paper_log_dir[0]){
		sprintf(master, "%s/%s", cfg->paper_log_dir, LATEST_MSG);
		len = readlink(master, file, PATH_MAX);
		if (len){
			file[len] = '\0';
			if ((startp = strrchr(file, '/')) != NULL){
				d_printf("link to %s\n", file);
				count = strtol(startp+1, &endp, 10);
			} else {
				count = strtol(file, &endp, 10);
			}
		}
	}

	/* ɸϤؤ¸*/
	count++;
	msg->serial_no = count;
	if (cfg->live){
		print_message(fp, msg, psaddr, ulm);
#if defined(ENDMARK)
		fprintf(fp, "\r\n");
#else
		fprintf(fp, "\n");
#endif
	}

	/* ǿեؤΥ */
    if (master[0]){
        sprintf(file, "%s/%0" SAVE_FILE_NUM_MAX "d", 
				cfg->paper_log_dir, count);
        if ((fp = fopen(file, "w"))){
            print_message(fp, msg, psaddr, ulm);
        
            fclose(fp);
            unlink(master);
            symlink(file, master);
        } else {
            fprintf(stderr, "Can't open save file(%s)", file);
            perror(NULL);
        }
    }

	/* եؤ¸ */
    if (cfg->scroll_log_file[0]){
        if ((fp = fopen(cfg->scroll_log_file, "a"))){
            print_message(fp, msg, psaddr, ulm);
            fclose(fp);
        } else {
            fprintf(stderr, "Can't open log file(%s)", 
					cfg->scroll_log_file);
            perror(NULL);
			return 0;
        }
    }

    return count;
}

/* åɽ */
int print_message(FILE* fp, MSG* msg, struct sockaddr_in* psaddr, 
				  USERLIST_MNG* ulm)
{
	USERLIST* ul;
    char user[NAMELEN];
    char host[NAMELEN];
	static char horizon[] = "\n----------------------------------------------\n";
/*	static char horizon_end[] = "----------------Message End-------------------\n\n"; */
	char buf[MAXBUFLEN];
	int lp;

    time_t now_time;
    time( &now_time );

	if ((ul = search_userlist_by_addr(ulm, psaddr->sin_addr.s_addr))){
		strcpy(user,ul->nickname[0] ? ul->nickname : ul->user);
		sprintf(host, "%s%s%s", 
				ul->groupname,
				ul->groupname[0] ? "/" : "", 
				ul->host);
	} else {
		strcpy(user, msg->msg_user);
		sprintf(host, "[%s]", msg->msg_host);
	}

/* 󥿥󥰴ֳַ׻ buf ɽ */
#define PM2CENTER(x, y, z) do {\
  if (z > strlen(y)){\
    for (lp=0;  lp < (z - strlen(y)) / 2; lp++){\
      fprintf(x, " ");\
    }\
  }\
  fprintf(x, y);\
}while(0)

	/* ɽ */
	fprintf(fp, "#%d", msg->serial_no);
    fprintf(fp, horizon);
    if (GET_OPT(msg->command) & IPMSG_BROADCASTOPT){
		/* ֥ɥ㥹 */
		sprintf(buf, "Broadcast from [%s] ( %s )\n", user, host);
		PM2CENTER(fp, buf, sizeof(horizon));
    } else {				
        if (GET_OPT(msg->command) & IPMSG_MULTICASTOPT){
			/* ޥ㥹 */
            sprintf(buf, "Multicast from [%s] ( %s )\n", user, host);
			PM2CENTER(fp, buf, sizeof(horizon));
        } else {
			/* ̾ */
            sprintf(buf, "Message from [%s] ( %s )\n", user, host );
			PM2CENTER(fp, buf, sizeof(horizon));
        }
    }
    sprintf(buf, "at %s", ctime(&now_time));
	PM2CENTER(fp, buf, sizeof(horizon));
    fprintf(fp, horizon);
    fprintf(fp, "%s\n", msg->extend);
/*    fprintf(fp, horizon_end); */

    return 1;
}

/* USER LIST ؤɲ */
int append_userlist(USERLIST_MNG* ulm, USERLIST* ul)
{
    delete_userlist_by_addr(ulm, ul->addr);

    d_printf("Append [%s]@[%s]\n", ul->user, ul->host);
    ul->next = NULL;
    ul->prev = ulm->last;

    if (ulm->last){
        ulm->last->next = ul;
    } else {
        d_printf("Insert first\n");
        ulm->first = ul;
    }
	ulm->last = ul;
    
    return 1;
}

/* 桼ꥹȥ󤫤κ
   ul ϥ桼ꥹ¸ߤƤ뤳 */
int unlink_userlist(USERLIST_MNG* mng, USERLIST* ul)
{
	if (ul->prev){
		ul->prev->next = ul->next;
	} else {
		mng->first = ul->next;
	}
	if (ul->next){
		ul->next->prev = ul->prev;
	} else {
		mng->last = ul->prev;
	}
	return TRUE;
}

/* ADDRESS ˤ USER LIST κ(Τκ) */
int delete_userlist_by_addr(USERLIST_MNG* mng, long addr)
{
    USERLIST* ul; /* Userlist Target*/
    if (!(ul = search_userlist_by_addr(mng, addr))){
        /* User no exist */
        return 0;
    }
	unlink_userlist(mng, ul);
    free(ul);
    return 1;
}

/* USER LIST ˴ */
void destroy_userlist(USERLIST_MNG* ulm)
{
    USERLIST* ult; /* Userlist Target*/

    d_printf("Destroy_userlist Start!\n");
    while((ult = ulm->first) != NULL){
        d_printf("Delete Userlist[%s]\n", ult->user);
        ulm->first = ult->next;
        free(ult);
    }
	ulm->last = NULL;
    d_printf("Destroy_userlist E n d!\n");
}

/* ɥ쥹ˤ USER LIST θ */
USERLIST* search_userlist_by_addr(USERLIST_MNG* mng, long addr)
{
	USERLIST* ul;

	if (!mng || !mng->first){
		return NULL;
	}

    for (ul = mng->first; ul; ul = ul->next){
        if (ul->addr == addr){
            d_printf("Search user by addr(%s)\n", ul->user);
			take_first_userlist(mng, ul);
            return ul;
        }
    }
    return NULL;
}

/* ۥ̾ˤ桼ꥹȤθ */
USERLIST* search_userlist_by_host(USERLIST_MNG* mng, char* host)
{
	USERLIST* ul;

	if (!mng || !mng->first){
		return NULL;
	}

    for (ul = mng->first; ul; ul = ul->next){
        if (strcasecmp(ul->host, host) == 0){
            d_printf("Search user by host(%s)\n", ul->user);
			take_first_userlist(mng, ul);
            return ul;
        }
    }
    return NULL;
}

/* 桼̾ˤ桼ꥹȤθ */
USERLIST* search_userlist_by_user(USERLIST_MNG* mng, char* user)
{
	USERLIST* ul;

	if (!mng || !mng->first){
		return NULL;
	}

    for (ul = mng->first; ul; ul = ul->next){
        if (strcasecmp(ul->nickname, user) == 0){
            d_printf("Search host by nickname(%s)\n", ul->host);
			take_first_userlist(mng, ul);
            return ul;
        }
    }
    for (ul = mng->first; ul; ul = ul->next){
        if (strcasecmp(ul->user, user) == 0){
            d_printf("Search host by user(%s)\n", ul->host);
			take_first_userlist(mng, ul);
            return ul;
        }
    }
    return NULL;
}

/* 桼ꥹȤƬ˰ư */
int take_first_userlist(USERLIST_MNG* mng, USERLIST* ul)
{
	if (ul == mng->first) return FALSE;

	unlink_userlist(mng, ul);
	ul->prev = NULL;
	ul->next = mng->first;
	insert_userlist(mng, ul);

	return TRUE;
}

/* 桼ꥹȤ
   (ul  prev, next ϻꤵƤ뤳) */
int insert_userlist(USERLIST_MNG* mng, USERLIST* ul)
{
	if (ul->prev){
		ul->prev->next = ul;
	} else {
		mng->first = ul;
	}
	if (ul->next){
		ul->next->prev = ul;
	} else {
		mng->last = ul;
	}
	return TRUE;
}

/* USERLIST 촹
    s1,s2  USERLIST ¸ߤƤ뤳 */
int swap_userlist(USERLIST_MNG* mng, USERLIST* s1, USERLIST* s2)
{
	USERLIST* tmp;
#if(0)
	if (!is_in_userlist(s1) || is_in_userlist(s2)){
		return FALSE;
	}
#endif
	tmp = s1->prev;
	s1->prev = s2->prev;
	s2->prev = tmp;

	tmp = s1->next;
	s1->next = s2->next;
	s2->next = tmp;

	insert_userlist(mng, s1);
	insert_userlist(mng, s2);

	return TRUE;
}

/* 桼ꥹ˻ ul ¸ߤ뤳Ȥγǧ */
int is_in_userlist(USERLIST_MNG* mng, USERLIST* ul)
{
	if (!ul) return FALSE;

	if (ul->prev){
		if (ul->prev->next != ul) return FALSE;
	} else {
		if (mng->first != ul) return FALSE;
	}
	if (ul->next){
		if (ul->next->prev != ul) return FALSE;
	} else {
		if (mng->last != ul) return FALSE;
	}
	
	return TRUE;
}

/*  USER LIST եǤФ */
int flush_userlist(CONFIG_MNG* cfg, USERLIST_MNG* ulm)
{
    USERLIST* ul;
    FILE* fp;
    FILE* fp_logging = NULL;
	
	int flag_flush = ulm->flag_flush;
	ulm->flag_flush=FALSE;
    if (!cfg->userlist_file[0] || !ulm->first){
		d_printf("No path or No first presented.\n");
        return 0;
    }

    d_printf("flush_userlist\n");

    if (!(fp = fopen(cfg->userlist_file, "w"))){
        fprintf(stderr, "Can't open userlist file(%s):", cfg->userlist_file);
        perror(NULL);
        return 0;
    }
	if (flag_flush && cfg->userlist_logging){
		if (!(fp_logging = fopen(cfg->userlist_logfile, "a"))){
			fprintf(stderr, "Can't open userlist log file(%s):",
					cfg->userlist_logfile);
			perror(NULL);
		} else {
			time_t t = time(NULL);
			fprintf(fp_logging, "---- UPDATE USERLIST ---- %s", ctime(&t));
		}
	} else {
		d_printf("flag_flush %d, logging %d\n", flag_flush, cfg->userlist_logging);
	}
    
    for (ul=ulm->first; ul; ul=ul->next){
		char buf[MAXBUFLEN];
		struct in_addr in;
		in.s_addr = ul->addr;
		switch(cfg->type_of_userlist){
		case TAB_USERLIST:
			snprintf(buf, MAXBUFLEN, "%s\t%s\t%s\t%s\t%s\t%c\n", 
					 ul->nickname[0] ? ul->nickname : ul->user,
					 ul->groupname, ul->host, ul->user, inet_ntoa(in),
					 ul->absence);
			break;
		case NORMAL_USERLIST:
			snprintf(buf, MAXBUFLEN, "%s@%s%s%s[%s](%s:%s:%c)\n", 
					 ul->nickname[0] ? ul->nickname : ul->user,
					 ul->groupname, ul->groupname[0] ? "/" : "", 
					 ul->host, inet_ntoa(in), ul->user, ul->host,
					 ul->absence);
		default:
			break;
		}
        fprintf(fp, "%s", buf);
		if (fp_logging) fprintf(fp_logging, "%s", buf);
        d_printf(buf);
    }
    fclose(fp);
    if (fp_logging) fclose(fp_logging);
    return 1;
}

/* MSG  USERLIST Ȥ */
USERLIST* construct_userlist_from_msg(MSG* msg, struct sockaddr_in* saddr)
{
	USERLIST* ul;
	if (!(ul = (USERLIST*)malloc(sizeof(USERLIST)))){
		perror("Can't allocate userlist buffer.");
		return NULL;
	}

	ul->next = ul->prev = NULL;

	ul->addr = saddr->sin_addr.s_addr;
	ul->port = saddr->sin_port;
	Xstrncpy(ul->user, msg->msg_user, NAMELEN);
	Xstrncpy(ul->host, msg->msg_host, NAMELEN);
	ul->absence = (GET_OPT(msg->command) & IPMSG_ABSENCEOPT) ? 
		ABSENCE_CHAR : UNABSENCE_CHAR;
	if (GET_MODE(msg->command) <= IPMSG_BR_ABSENCE){
		Xstrncpy(ul->nickname, msg->extend, NAMELEN);
		Xstrncpy(ul->groupname, msg->group, NAMELEN);
	} else {
		Xstrncpy(ul->nickname, msg->msg_user, NAMELEN);
		ul->groupname[0] = '\0';
	}
	return ul;
}

/* 
   åꥹѤΥХåեѰդ롣 
    ˴ˤ destruct_msgpackage 뤳(ľ free ʤ褦)
*/
MSGPACKAGE* construct_msgpackage(USERLIST* ul, const MSG* msg)
{
	MSGPACKAGE* package;
	char* content = msg->extend;
	int len;

	if (!content || !(len=strlen(content))){
		d_printf("NULL buffer presented create_msgpackage.\n");
		return NULL;
	}

	if (!(package = (MSGPACKAGE*)calloc(1, sizeof(MSGPACKAGE)))){
		perror("Can't allocate message buffer.");
		return NULL;
	}

	package->prev = package->next = NULL;
	package->serial_no = msg->serial_no;
	package->received_time = time(NULL);
	package->len = msg->len;

	package->user = strdup(ul->user);
	package->host = strdup(ul->host);
	package->nickname = strdup(ul->nickname);
	package->groupname = strdup(ul->groupname);
	package->content = strdup(content);
	d_printf("NICKNAME is [%s]\n", package->nickname);
	if (package->user && package->host && package->nickname &&
		package->groupname && package->content){
		return package;
	}

	destruct_msgpackage(package);
	return NULL;
}

void destruct_msgpackage(MSGPACKAGE* package)
{
	if (!package) return;
	if (package->user) free(package->user);
	if (package->host) free(package->host);
	if (package->nickname) free(package->nickname);
	if (package->groupname) free(package->groupname);
	if (package->content) free(package->content);
	free(package);
}

int append_msgpackage(MSGPACKAGE_MNG* mng, MSGPACKAGE* package)
{
	package->prev = mng->last;
	package->next = NULL;

	if (!mng->last){
		mng->first = mng->current = package;
	} else {
		package->prev->next = package;
	}

	mng->last = package;
	if (mng->store_message >= mng->max_store_message){
		delete_msgpackage(mng, mng->first);
	} else {
		mng->store_message++;
	}
	return 1;
}

MSGPACKAGE* search_msgpackage_by_serial_no(MSGPACKAGE_MNG* mng, int number)
{
	int forward;
	MSGPACKAGE* ret;

	if (!mng || !mng->current || number < 0){
		return NULL;
	}

	forward = (mng->current->serial_no < number) ? TRUE : FALSE;
	for (ret = mng->current; ret; ret = forward ? ret->next : ret->prev){
		if (ret->serial_no == number){
			break;
		}
	}

	return ret;
}

int delete_msgpackage_by_serial_no(MSGPACKAGE_MNG* mng, int number)
{
	MSGPACKAGE* target;

	if (!(target = search_msgpackage_by_serial_no(mng, number))){
		return FALSE;
	}
	return delete_msgpackage(mng, target);
}
	
int delete_msgpackage(MSGPACKAGE_MNG* mng, MSGPACKAGE* target)
{
	if (target->prev){
		target->prev->next = target->next;
	} else {
		mng->first = target->next;
	}
	if (target->next){
		target->next->prev = target->prev;
	} else {
		mng->last = target->prev;
	}

	if (target == mng->current){
		if (target->next){
			mng->current = target->next;
		} else {
			mng->current = target->prev;
		}
	}
	destruct_msgpackage(target);
	return TRUE;
}

MSGPACKAGE* setforward_msgpackage_current(MSGPACKAGE_MNG* mng)
{
	if (!mng || !mng->current){
		return NULL;
	}

	if (mng->current->next){
		mng->current = mng->current->next;
		return mng->current;
	}
	return NULL;
}

MSGPACKAGE* setbackward_msgpackage_current(MSGPACKAGE_MNG* mng)
{
	if (!mng || !mng->current){
		return NULL;
	}

	if (mng->current->prev){
		mng->current = mng->current->prev;
		return mng->current;
	}
	return NULL;
}

MSGPACKAGE* set_msgpackage_current(MSGPACKAGE_MNG* mng, int number)
{
	MSGPACKAGE* ret;
	if (!mng || !mng->current || number < 0){
		return NULL;
	}
	if (!(ret = search_msgpackage_by_serial_no(mng, number))){
		return NULL;
	}
	return mng->current = ret;
}

int serialize_msgpackage(char* buf, int size, MSGPACKAGE* msg)
{
	int len;
	len = sprintf(buf, "%d:%ld:%s%c%s%c%s%c%s%c", msg->serial_no, 
				  msg->received_time, msg->user, 0, msg->host, 0,
				  msg->nickname, 0, msg->groupname, 0);
	len += Xstrncpy(buf + len, msg->content, size - len);
	return len;
}

MSGPACKAGE* de_serialize_msgpackage(char* buf, int len, MSGPACKAGE* msg)
{
	char* endp;
	msg->serial_no = strtol(buf, &endp, 0);
	msg->received_time = strtol(endp+1, &endp, 0);

	msg->user = strdup(endp+1);
	endp += strlen(msg->user);
	if (endp - buf > len) return NULL;

	msg->host = strdup(endp+1);
	endp += strlen(msg->host);
	if (endp - buf > len) return NULL;

	msg->nickname = strdup(endp+1);
	endp += strlen(msg->nickname);
	if (endp - buf > len) return NULL;

	msg->groupname = strdup(endp+1);
	endp += strlen(msg->groupname);
	if (endp - buf > len) return NULL;

	msg->content = strdup(endp+1);
	return msg;
}

/* 
   ǡ⡼ɤؤΤ
   Ʊե뤬ٵưƵư줿ץǤ
   chdir Τߤ򤪤ʤäƤ롣
 */
void became_a_daemon(char* argv[])
{
	pid_t pid;
	if (*argv[0] == '['){
		if (chdir("/") < 0){
			perror("chdir faild");
		}
		return;
	}
	if ((pid = fork())<0){
		perror("fork faild");
		return;
	} else if (pid > 0){
		wait(NULL);
		exit(0);
	}

	if (setsid() < 0){
		perror("setsid faild");
	}

	if ((pid = fork()) == 0){
		char* prog;
		prog = argv[0];
		argv[0] = strdup("[tipmsg]");
		execvp(prog, argv);
		perror("execvp fail");
		fprintf(stderr, "Prog[%s]\n", prog);
		_exit(1);
	}
	_exit(0);
}

void code_transferSE(unsigned char* first, unsigned char* second)
{
	*first <<= 1;
	if( *second < 0x9F ){
		if(*first < 0x3F) *first -= 0x61; else *first += 0x1F;
		if(*second > 0x7E) *second += 0x60; else *second += 0x61;
	} else {
		if(*first < 0x3F) *first -= 0x60; else *first += 0x20;
		*second += 0x02;
	}
}

void code_transferES(unsigned char* first, unsigned char* second)
{
	if (*first & 0x01){
		*first >>= 1;
		if (*first < 0x6F) *first += 0x31; else *first += 0x71;
		if (*second > 0xDF) *second -= 0x60; else *second -= 0x61;
	} else {
		*first >>= 1;
		if (*first < 0x6F) *first += 0x30; else *first += 0x70;
		*second -= 0x02;
	}
}

int XSJIS_htof(CONFIG_MNG* cfg, unsigned char* msg, int pos, int* len, int buflen)
{
	if (msg[pos] > 0xA0 && msg[pos] < 0xE0) {
		if (cfg->hw_kana == 1){
			if (msg[pos+1] > 0xA0 && msg[pos+1] < 0xE0) {
				SJIS_htof(msg + pos, msg + pos + 1);
			} else {
				msg[pos] = 'x';
				return -1;
			}
		} else if (cfg->hw_kana == 2){
			if (++(*len) >= buflen) return *len;
			memmove(msg + pos +2, msg + pos +1, *len -pos -2);
			SJIS_htof(msg + pos, msg + pos + 1);
		} else if (cfg->hw_kana == 3){
			if (++(*len) >= buflen) return *len;
			memmove(msg + pos +2, msg + pos +1, *len -pos -2);
			msg[pos+1] = msg[pos] - 0x80; msg[pos]=0x8E;
			return -1;
		} else {
			msg[pos] = 'x';
			return -1;
		}
	}
	return *len;
}

void SJIS_htof(unsigned char* first, unsigned char* second)
{
	int index;
	static unsigned char table[] = {
		0x81,0x42,0x81,0x75,0x81,0x76,0x81,0x41,
		0x81,0x45,0x83,0x92,0x83,0x40,0x83,0x42,
		0x83,0x44,0x83,0x46,0x83,0x48,0x83,0x83,
		0x83,0x85,0x83,0x87,0x83,0x62,0x81,0x5B,
		0x83,0x41,0x83,0x43,0x83,0x45,0x83,0x47,
		0x83,0x49,0x83,0x4A,0x83,0x4C,0x83,0x4E,
		0x83,0x50,0x83,0x52,0x83,0x54,0x83,0x56,
		0x83,0x58,0x83,0x5A,0x83,0x5C,0x83,0x5E,
		0x83,0x60,0x83,0x63,0x83,0x65,0x83,0x67,
		0x83,0x69,0x83,0x6A,0x83,0x6B,0x83,0x6C,
		0x83,0x6D,0x83,0x6E,0x83,0x71,0x83,0x74,
		0x83,0x77,0x83,0x7A,0x83,0x7D,0x83,0x7E,
		0x83,0x80,0x83,0x81,0x83,0x82,0x83,0x84,
		0x83,0x86,0x83,0x88,0x83,0x89,0x83,0x8A,
		0x83,0x8B,0x83,0x8C,0x83,0x8D,0x83,0x8F,
		0x83,0x93,0x81,0x4A,0x81,0x4B };

	*second = 'x';
	if(*first < 0xA1 || *second > 0xE0) return;

	index = (*first - 0xA1) *2;
	*first = table[index+0];
	*second = table[index+1];
}

int Xstrncpy(char* dest, const char* src, int len)
{
	int cnt;
	for (cnt = 0;  cnt < len-1; cnt++){
		dest[cnt] = src[cnt];
		if (src[cnt] == '\0') return cnt;
	}
	dest[cnt]='\0';
	return cnt;
}

int Xselect(int fd, int timeout)
{
    fd_set rfds;
    struct timeval tv;
	int ret;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	if ((ret = select(fd + 1 , &rfds, NULL, NULL, &tv))<0){
		if (errno != EINTR) perror("select fail");
	}
	return ret;
}

int Xsendto(CONFIG_MNG* cfg, int s, const void *msg, size_t len, int flags,
       const struct sockaddr *to, int tolen)
{
	int sendlen = 0;
	/* Transfer euc to sjis */
	if (!cfg->char_code[0]) len = sendto(s, msg, len, flags, to, tolen);
	else {
		transfer_kanji_code(cfg, (unsigned char*)msg, &len, UDP_BUF, 
							cfg->char_code, CODE_SJIS);
		sendlen = sendto(s, msg, len, flags, to, tolen);
	}

	return sendlen;
}
	
int Xrecvfrom(CONFIG_MNG* cfg, int s, void  *buf, size_t len, int  flags,
       struct sockaddr *from, int *fromlen)
{
	int recvlen;
	recvlen = recvfrom(s, buf, len, flags, from, fromlen);
	
	if (!cfg->char_code[0]) return recvlen;
	transfer_kanji_code(cfg, (unsigned char*)buf, &recvlen, len, 
						CODE_SJIS, cfg->char_code);
	return recvlen;
}

/*  Now supported by ...

    SJIS  -> EUCJP
    EUCJP -> SJIS
    JIS   -> EUCJP
    JIS   -> SJIS
 */
char* transfer_kanji_code(CONFIG_MNG* cfg, unsigned char* buf, 
						  int* len, int size,
						  char* in, char* out)
{
	int lp, ret;
	if (strcasecmp(in, out) == 0) return buf;
	
	if (strcasecmp(in, CODE_SJIS) == 0){
		if (strcasecmp(out, CODE_EUCJP) == 0){
			for (lp=0; lp < *len; lp++) {
				if (buf[lp] < 0x81 || buf[lp] == 0xA0) continue;
				/* ȾѲ̾Ф */
				if ((ret = XSJIS_htof(cfg, buf, lp, len, size))<0) continue;
				if ((*len = ret) >= size) break;
				if (++lp >= *len) break;
				code_transferSE(buf+lp-1, buf+lp);
			}
		}
	} else if (strcasecmp(in, CODE_EUCJP) == 0){
		if (strcasecmp(out, CODE_SJIS) == 0){
			for (lp=0; lp < *len; lp++) {
				if (buf[lp] < 0x80) continue;
				if (++lp >= *len) break;
				code_transferES(buf + lp-1, buf + lp);
			}
		}
	} else if (strcasecmp(in, CODE_JIS) == 0){
		if (strcasecmp(out, CODE_EUCJP) == 0){
			char* p = buf;
			int pos = 0;
			int mode = ASCII;
			for (lp=0; lp < *len; lp++) {
				if((p[lp]=='\x0f') || (p[lp]=='\x0e') ||
				   (p[lp]=='\x1b' && p[lp+1]=='(' && p[lp+2] == 'B') ||
				   (p[lp]=='\x1b' && p[lp+1]=='(' && p[lp+2] == 'J') ||
				   (p[lp]=='\x1b' && p[lp+1]=='(' && p[lp+2] == 'I')){
					lp += 2;
					mode = ASCII;
					continue;
				} else if((p[lp]=='\x1b' && p[lp+1]=='$' && p[lp+2] == 'B') ||
						  (p[lp]=='\x1b' && p[lp+1]=='$' && p[lp+2] == '@')){
					lp += 2;
					mode = JAPANESE;
					continue;
				}
				if (mode == ASCII) p[pos++] = p[lp];
				else p[pos++] = p[lp] | 0x80;
			}
			p[pos] ='\0';
		} else if (strcasecmp(out, CODE_SJIS) == 0) {
			transfer_kanji_code(cfg, buf, len, size, in, CODE_EUCJP);
			transfer_kanji_code(cfg, buf, len, size, CODE_EUCJP, out);
		}
	}
	return buf;
}

/*  pid ե뤫ͤɤ߹ pidȤ֤ */
pid_t read_pid_file(const char* file)
{
	int fd;
	char pid[0x10];
	if ((fd = open(file, O_RDONLY)) == -1){
		fprintf(stderr, "Can't open pid file(%s) : %s\n", 
				file, strerror(errno));
		return 0;
	}
	
	if (read(fd, pid, 0x10) == -1){
		perror("Can't read pid from file.");
		close(fd);
		return 0;
	}
	close(fd);
	return strtol(pid, NULL, 0);
}

/* pid ե벽 */
int create_pid_file(const char* file)
{
	int fd, len;
	char pid[0x10];
	if ((fd = open(file, O_CREAT|O_TRUNC|O_WRONLY, 0600)) == -1){
		perror("Can't create pid file.");
		goto ERR0;
	}
	len = sprintf(pid, "%d", (int)getpid());
	
	if ((write(fd, pid, len)<len)){
		perror("Can't write pid to file.");
		goto ERR1;
	}
	close(fd);
	return TRUE;

  ERR1:
	close(fd);
  ERR0:
	return FALSE;
}

char* Xinet_ntoa(unsigned long addr)
{
	struct in_addr iaddr;
	char* p;
	iaddr.s_addr = addr;
	if (!(p = inet_ntoa(iaddr))) return "ERR";
	return p;
}

int sendACK(NET_MNG* net, CONFIG_MNG* cfg, struct sockaddr_in* pSaddr, int num)
{
	char buf[0x100];
	int len;
	d_printf("send to %s\n", inet_ntoa(pSaddr->sin_addr));
	len = sprintf(buf, "%d:%s:%d:", TIPMSG_CMD_VERSION, ACK, num);
	return sendto(net->sd, buf, len, 0, (struct sockaddr *)pSaddr, sizeof(*pSaddr));
}

int receiveACK(NET_MNG* net, CONFIG_MNG* cfg, int num)
{
	char* endp;
	int len, i;
	struct sockaddr_in recv;
	int	size = sizeof(struct sockaddr_in);
	char buf[0x100];

	if (Xselect(net->sd, 1)<=0) {
		d_printf("No reply\n");
		return 0;
	}
	if ((len = recvfrom(net->sd, buf, 0x100, 0, 
						(struct sockaddr *)&recv, &size))<=0){
		return len;
	}
	buf[len]='\0';
	d_printf("RECV[%s]\n", buf);
	if ((i = strtol(buf, &endp, 0)) != TIPMSG_CMD_VERSION){
		d_printf("Not TIPMSG cmd\n");
		return 0;
	}
	if (strncmp(endp+1, ACK, sizeof(ACK)-1) != 0){
		d_printf("Not ACK cmd(%s:%s) of %d\n", endp+1, ACK, sizeof(ACK)-1);
		return 0;
	}
	endp += sizeof(ACK) + 1;
	if ((i = strtol(endp, &endp, 0)) != num){
		d_printf("Different serial number(%d <-> %d.\n", i, num);
		return 0;
	}

	return len;
}

/* Τ˴ؤ */

/* ¤Τ䤤碌 */
static READMSG_MNG* query_for_readmsg_mng(void)
{
	static READMSG_MNG mng;
	return &mng;
}

/* READMSG ȯԤƤʤۥȤꥹȤɲ */
int add_for_readmsg_mng(long serial_no, long packet_no, long addr)
{
	READMSG_MNG* mng = query_for_readmsg_mng();
	READMSG_PKG* pkg;

	d_printf("%s", "add_noreply_readmsg.\n");
	if (!(pkg = (READMSG_PKG*)malloc(sizeof(READMSG_PKG)))){
		fprintf(stderr, "Can't allocate memory.");
		return 0;
	}

	pkg->packet_no = packet_no;
	pkg->serial_no = serial_no;
	pkg->addr = addr;

	if (!mng->last){
		mng->first = mng->last = pkg;
		pkg->next = pkg->prev = NULL;
	} else {
		mng->last->next = pkg;
		pkg->prev = mng->last;
		pkg->next = NULL;
		mng->last = pkg;
	}

	return 1;
}

/* READMSG ȯԤƤʤۥȤФ READMSG ȯ */
int request_for_readmsg_mng(int serial_no,
						   NET_MNG* net, CONFIG_MNG* cfg, USERLIST_MNG* ulm)
{
	READMSG_MNG* mng = query_for_readmsg_mng();
	READMSG_PKG* pkg;
	char buf[0x10];

	d_printf("request_for_readmsg_mng.\n");
	if (serial_no == 0){
		d_printf("Request READMSG for the latest message.\n");
		pkg = mng->last;
	} else {
		for (pkg = mng->first; pkg; pkg = pkg->next){
			if (pkg->serial_no == serial_no) break;
		}
	}
	if (!pkg) return 0;

	sprintf(buf, "%ld", pkg->packet_no);
	sender(buf, strlen(buf)+1, pkg->addr, IPMSG_READMSG, net, cfg, ulm);

	if (pkg->prev) pkg->prev->next = pkg->next;
	else mng->first = pkg->next;
	if (pkg->next) pkg->next->prev = pkg->prev;
	else mng->last = pkg->prev;
	free(pkg);

	return 1;
}

/*  ̤ READMSG 򤹤٤ȯ */
int request_all_for_readmsg_mng(
	NET_MNG* net, CONFIG_MNG* cfg, USERLIST_MNG* ulm)
{
	READMSG_MNG* mng = query_for_readmsg_mng();
	READMSG_PKG* pkg;
	READMSG_PKG* next;
	char buf[0x20];

	d_printf("request_all_for_readmsg_mng.\n");
	if (!mng->first) return 0;

	for (pkg = mng->first; pkg; pkg = next){
		d_printf("Serial no : %ld, Packet_no : %ld, Address : 0x%lx\n", 
				 pkg->serial_no, pkg->packet_no, pkg->addr);
		sprintf(buf, "%ld", pkg->packet_no);
		sender(buf, strlen(buf)+1, pkg->addr, IPMSG_READMSG|IPMSG_AUTORETOPT, 
			   net, cfg, ulm);
		next = pkg->next;
		free(pkg);
	}
	mng->first = mng->last = NULL;
	d_printf("Delete pkg complete.\n");

	return 1;
}

void dummy_function(char *dummy,...)
{
}
